@@ -9,61 +9,45 @@
 
 #define USE_XHTML(opt) (opt->flags & HOEDOWN_HTML_USE_XHTML)
 
-struct rndr_state {
-	struct {
-		int header_count;
-		int current_level;
-		int level_offset;
-		int nesting_level;
-	} toc_data;
-
-	unsigned int flags;
-
-	/* extra callbacks */
-	void (*link_attributes)(hoedown_buffer *ob, const hoedown_buffer *url, void *self);
-};
-
-typedef struct rndr_state rndr_state;
-
-int
-hoedown_html_is_tag(const uint8_t *tag_data, size_t tag_size, const char *tagname)
+hoedown_html_tag
+hoedown_html_is_tag(const uint8_t *data, size_t size, const char *tagname)
 {
 	size_t i;
 	int closed = 0;
 
-	if (tag_size < 3 || tag_data[0] != '<')
+	if (size < 3 || data[0] != '<')
 		return HOEDOWN_HTML_TAG_NONE;
 
 	i = 1;
 
-	if (tag_data[i] == '/') {
+	if (data[i] == '/') {
 		closed = 1;
 		i++;
 	}
 
-	for (; i < tag_size; ++i, ++tagname) {
+	for (; i < size; ++i, ++tagname) {
 		if (*tagname == 0)
 			break;
 
-		if (tag_data[i] != *tagname)
+		if (data[i] != *tagname)
 			return HOEDOWN_HTML_TAG_NONE;
 	}
 
-	if (i == tag_size)
+	if (i == size)
 		return HOEDOWN_HTML_TAG_NONE;
 
-	if (isspace(tag_data[i]) || tag_data[i] == '>')
+	if (isspace(data[i]) || data[i] == '>')
 		return closed ? HOEDOWN_HTML_TAG_CLOSE : HOEDOWN_HTML_TAG_OPEN;
 
 	return HOEDOWN_HTML_TAG_NONE;
 }
 
-static inline void escape_html(hoedown_buffer *ob, const uint8_t *source, size_t length)
+static void escape_html(hoedown_buffer *ob, const uint8_t *source, size_t length)
 {
 	hoedown_escape_html(ob, source, length, 0);
 }
 
-static inline void escape_href(hoedown_buffer *ob, const uint8_t *source, size_t length)
+static void escape_href(hoedown_buffer *ob, const uint8_t *source, size_t length)
 {
 	hoedown_escape_href(ob, source, length);
 }
@@ -72,18 +56,13 @@ static inline void escape_href(hoedown_buffer *ob, const uint8_t *source, size_t
  * GENERIC RENDERER *
  ********************/
 static int
-rndr_autolink(hoedown_buffer *ob, const hoedown_buffer *link, enum hoedown_autolink type, void *opaque)
+rndr_autolink(hoedown_buffer *ob, const hoedown_buffer *link, hoedown_autolink_type type, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 
 	if (!link || !link->size)
 		return 0;
 
-	if ((state->flags & HOEDOWN_HTML_SAFELINK) != 0 &&
-		!hoedown_autolink_is_safe(link->data, link->size) &&
-		type != HOEDOWN_AUTOLINK_EMAIL)
-		return 0;
-
 	HOEDOWN_BUFPUTSL(ob, "<a href=\"");
 	if (type == HOEDOWN_AUTOLINK_EMAIL)
 		HOEDOWN_BUFPUTSL(ob, "mailto:");
@@ -91,7 +70,7 @@ rndr_autolink(hoedown_buffer *ob, const hoedown_buffer *link, enum hoedown_autol
 
 	if (state->link_attributes) {
 		hoedown_buffer_putc(ob, '\"');
-		state->link_attributes(ob, link, opaque);
+		state->link_attributes(ob, link, data);
 		hoedown_buffer_putc(ob, '>');
 	} else {
 		HOEDOWN_BUFPUTSL(ob, "\">");
@@ -114,41 +93,14 @@ rndr_autolink(hoedown_buffer *ob, const hoedown_buffer *link, enum hoedown_autol
 }
 
 static void
-rndr_blockcode(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_buffer *lang, void *opaque)
+rndr_blockcode(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_buffer *lang, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
-
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
 
-	if (lang && lang->size) {
-		size_t i, cls = 0;
-		if (state->flags & HOEDOWN_HTML_PRETTIFY) {
-			HOEDOWN_BUFPUTSL(ob, "<pre><code class=\"prettyprint");
-			cls++;
-		} else {
-			HOEDOWN_BUFPUTSL(ob, "<pre><code class=\"");
-		}
-
-		for (i = 0; i < lang->size; ++i, ++cls) {
-			while (i < lang->size && isspace(lang->data[i]))
-				i++;
-
-			if (i < lang->size) {
-				size_t org = i;
-				while (i < lang->size && !isspace(lang->data[i]))
-					i++;
-
-				if (lang->data[org] == '.')
-					org++;
-
-				if (cls) hoedown_buffer_putc(ob, ' ');
-				escape_html(ob, lang->data + org, i - org);
-			}
-		}
-
+	if (lang) {
+		HOEDOWN_BUFPUTSL(ob, "<pre><code class=\"language-");
+		escape_html(ob, lang->data, lang->size);
 		HOEDOWN_BUFPUTSL(ob, "\">");
-	} else if (state->flags & HOEDOWN_HTML_PRETTIFY) {
-		HOEDOWN_BUFPUTSL(ob, "<pre><code class=\"prettyprint\">");
 	} else {
 		HOEDOWN_BUFPUTSL(ob, "<pre><code>");
 	}
@@ -160,133 +112,126 @@ rndr_blockcode(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_buf
 }
 
 static void
-rndr_blockquote(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_blockquote(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
 	HOEDOWN_BUFPUTSL(ob, "<blockquote>\n");
-	if (text) hoedown_buffer_put(ob, text->data, text->size);
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</blockquote>\n");
 }
 
 static int
-rndr_codespan(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_codespan(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
-	if (state->flags & HOEDOWN_HTML_PRETTIFY)
-		HOEDOWN_BUFPUTSL(ob, "<code class=\"prettyprint\">");
-	else
-		HOEDOWN_BUFPUTSL(ob, "<code>");
+	HOEDOWN_BUFPUTSL(ob, "<code>");
 	if (text) escape_html(ob, text->data, text->size);
 	HOEDOWN_BUFPUTSL(ob, "</code>");
 	return 1;
 }
 
 static int
-rndr_strikethrough(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_strikethrough(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<del>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</del>");
 	return 1;
 }
 
 static int
-rndr_double_emphasis(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_double_emphasis(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<strong>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</strong>");
 
 	return 1;
 }
 
 static int
-rndr_emphasis(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_emphasis(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size) return 0;
+	if (!content || !content->size) return 0;
 	HOEDOWN_BUFPUTSL(ob, "<em>");
-	if (text) hoedown_buffer_put(ob, text->data, text->size);
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</em>");
 	return 1;
 }
 
 static int
-rndr_underline(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_underline(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<u>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</u>");
 
 	return 1;
 }
 
 static int
-rndr_highlight(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_highlight(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<mark>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</mark>");
 
 	return 1;
 }
 
 static int
-rndr_quote(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_quote(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<q>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</q>");
 
 	return 1;
 }
 
 static int
-rndr_linebreak(hoedown_buffer *ob, void *opaque)
+rndr_linebreak(hoedown_buffer *ob, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 	hoedown_buffer_puts(ob, USE_XHTML(state) ? "<br/>\n" : "<br>\n");
 	return 1;
 }
 
 static void
-rndr_header(hoedown_buffer *ob, const hoedown_buffer *text, int level, void *opaque)
+rndr_header(hoedown_buffer *ob, const hoedown_buffer *content, int level, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 
 	if (ob->size)
 		hoedown_buffer_putc(ob, '\n');
 
-	if ((state->flags & HOEDOWN_HTML_TOC) && (level <= state->toc_data.nesting_level))
+	if (level <= state->toc_data.nesting_level)
 		hoedown_buffer_printf(ob, "<h%d id=\"toc_%d\">", level, state->toc_data.header_count++);
 	else
 		hoedown_buffer_printf(ob, "<h%d>", level);
 
-	if (text) hoedown_buffer_put(ob, text->data, text->size);
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
 	hoedown_buffer_printf(ob, "</h%d>\n", level);
 }
 
 static int
-rndr_link(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_buffer *content, void *opaque)
+rndr_link(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
-
-	if (link != NULL && (state->flags & HOEDOWN_HTML_SAFELINK) != 0 && !hoedown_autolink_is_safe(link->data, link->size))
-		return 0;
+	hoedown_html_renderer_state *state = data->opaque;
 
 	HOEDOWN_BUFPUTSL(ob, "<a href=\"");
 
@@ -300,7 +245,7 @@ rndr_link(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *
 
 	if (state->link_attributes) {
 		hoedown_buffer_putc(ob, '\"');
-		state->link_attributes(ob, link, opaque);
+		state->link_attributes(ob, link, data);
 		hoedown_buffer_putc(ob, '>');
 	} else {
 		HOEDOWN_BUFPUTSL(ob, "\">");
@@ -312,108 +257,120 @@ rndr_link(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *
 }
 
 static void
-rndr_list(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *opaque)
+rndr_list(hoedown_buffer *ob, const hoedown_buffer *content, hoedown_list_flags flags, const hoedown_renderer_data *data)
 {
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
-	hoedown_buffer_put(ob, flags & HOEDOWN_LIST_ORDERED ? "<ol>\n" : "<ul>\n", 5);
-	if (text) hoedown_buffer_put(ob, text->data, text->size);
-	hoedown_buffer_put(ob, flags & HOEDOWN_LIST_ORDERED ? "</ol>\n" : "</ul>\n", 6);
+	hoedown_buffer_put(ob, (const uint8_t *)(flags & HOEDOWN_LIST_ORDERED ? "<ol>\n" : "<ul>\n"), 5);
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
+	hoedown_buffer_put(ob, (const uint8_t *)(flags & HOEDOWN_LIST_ORDERED ? "</ol>\n" : "</ul>\n"), 6);
 }
 
 static void
-rndr_listitem(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *opaque)
+rndr_listitem(hoedown_buffer *ob, const hoedown_buffer *content, hoedown_list_flags flags, const hoedown_renderer_data *data)
 {
 	HOEDOWN_BUFPUTSL(ob, "<li>");
-	if (text) {
-		size_t size = text->size;
-		while (size && text->data[size - 1] == '\n')
+	if (content) {
+		size_t size = content->size;
+		while (size && content->data[size - 1] == '\n')
 			size--;
 
-		hoedown_buffer_put(ob, text->data, size);
+		hoedown_buffer_put(ob, content->data, size);
 	}
 	HOEDOWN_BUFPUTSL(ob, "</li>\n");
 }
 
 static void
-rndr_paragraph(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_paragraph(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 	size_t i = 0;
 
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
 
-	if (!text || !text->size)
+	if (!content || !content->size)
 		return;
 
-	while (i < text->size && isspace(text->data[i])) i++;
+	while (i < content->size && isspace(content->data[i])) i++;
 
-	if (i == text->size)
+	if (i == content->size)
 		return;
 
 	HOEDOWN_BUFPUTSL(ob, "<p>");
 	if (state->flags & HOEDOWN_HTML_HARD_WRAP) {
 		size_t org;
-		while (i < text->size) {
+		while (i < content->size) {
 			org = i;
-			while (i < text->size && text->data[i] != '\n')
+			while (i < content->size && content->data[i] != '\n')
 				i++;
 
 			if (i > org)
-				hoedown_buffer_put(ob, text->data + org, i - org);
+				hoedown_buffer_put(ob, content->data + org, i - org);
 
 			/*
 			 * do not insert a line break if this newline
 			 * is the last character on the paragraph
 			 */
-			if (i >= text->size - 1)
+			if (i >= content->size - 1)
 				break;
 
-			rndr_linebreak(ob, opaque);
+			rndr_linebreak(ob, data);
 			i++;
 		}
 	} else {
-		hoedown_buffer_put(ob, &text->data[i], text->size - i);
+		hoedown_buffer_put(ob, content->data + i, content->size - i);
 	}
 	HOEDOWN_BUFPUTSL(ob, "</p>\n");
 }
 
 static void
-rndr_raw_block(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_raw_block(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_renderer_data *data)
 {
 	size_t org, sz;
-	if (!text) return;
+
+	if (!text)
+		return;
+
+	/* FIXME: Do we *really* need to trim the HTML? How does that make a difference? */
 	sz = text->size;
-	while (sz > 0 && text->data[sz - 1] == '\n') sz--;
+	while (sz > 0 && text->data[sz - 1] == '\n')
+		sz--;
+
 	org = 0;
-	while (org < sz && text->data[org] == '\n') org++;
-	if (org >= sz) return;
-	if (ob->size) hoedown_buffer_putc(ob, '\n');
+	while (org < sz && text->data[org] == '\n')
+		org++;
+
+	if (org >= sz)
+		return;
+
+	if (ob->size)
+		hoedown_buffer_putc(ob, '\n');
+
 	hoedown_buffer_put(ob, text->data + org, sz - org);
 	hoedown_buffer_putc(ob, '\n');
 }
 
 static int
-rndr_triple_emphasis(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_triple_emphasis(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size) return 0;
+	if (!content || !content->size) return 0;
 	HOEDOWN_BUFPUTSL(ob, "<strong><em>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</em></strong>");
 	return 1;
 }
 
 static void
-rndr_hrule(hoedown_buffer *ob, void *opaque)
+rndr_hrule(hoedown_buffer *ob, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
 	hoedown_buffer_puts(ob, USE_XHTML(state) ? "<hr/>\n" : "<hr>\n");
 }
 
 static int
-rndr_image(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_buffer *alt, void *opaque)
+rndr_image(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_buffer *alt, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 	if (!link || !link->size) return 0;
 
 	HOEDOWN_BUFPUTSL(ob, "<img src=\"");
@@ -432,12 +389,12 @@ rndr_image(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer
 }
 
 static int
-rndr_raw_html(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_raw_html(hoedown_buffer *ob, const hoedown_buffer *text, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 
-	/* HTML_ESCAPE overrides SKIP_HTML, SKIP_STYLE, SKIP_LINKS and SKIP_IMAGES
-	* It doens't see if there are any valid tags, just escape all of them. */
+	/* ESCAPE overrides SKIP_HTML. It doesn't look to see if
+	 * there are any valid tags, just escapes all of them. */
 	if((state->flags & HOEDOWN_HTML_ESCAPE) != 0) {
 		escape_html(ob, text->data, text->size);
 		return 1;
@@ -446,46 +403,47 @@ rndr_raw_html(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
 	if ((state->flags & HOEDOWN_HTML_SKIP_HTML) != 0)
 		return 1;
 
-	if ((state->flags & HOEDOWN_HTML_SKIP_STYLE) != 0 &&
-		hoedown_html_is_tag(text->data, text->size, "style"))
-		return 1;
-
-	if ((state->flags & HOEDOWN_HTML_SKIP_LINKS) != 0 &&
-		hoedown_html_is_tag(text->data, text->size, "a"))
-		return 1;
-
-	if ((state->flags & HOEDOWN_HTML_SKIP_IMAGES) != 0 &&
-		hoedown_html_is_tag(text->data, text->size, "img"))
-		return 1;
-
 	hoedown_buffer_put(ob, text->data, text->size);
 	return 1;
 }
 
 static void
-rndr_table(hoedown_buffer *ob, const hoedown_buffer *header, const hoedown_buffer *body, void *opaque)
+rndr_table(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (ob->size) hoedown_buffer_putc(ob, '\n');
-	HOEDOWN_BUFPUTSL(ob, "<table><thead>\n");
-	if (header)
-		hoedown_buffer_put(ob, header->data, header->size);
-	HOEDOWN_BUFPUTSL(ob, "</thead><tbody>\n");
-	if (body)
-		hoedown_buffer_put(ob, body->data, body->size);
-	HOEDOWN_BUFPUTSL(ob, "</tbody></table>\n");
+    if (ob->size) hoedown_buffer_putc(ob, '\n');
+    HOEDOWN_BUFPUTSL(ob, "<table>\n");
+    hoedown_buffer_put(ob, content->data, content->size);
+    HOEDOWN_BUFPUTSL(ob, "</table>\n");
 }
 
 static void
-rndr_tablerow(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_table_header(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
+{
+    if (ob->size) hoedown_buffer_putc(ob, '\n');
+    HOEDOWN_BUFPUTSL(ob, "<thead>\n");
+    hoedown_buffer_put(ob, content->data, content->size);
+    HOEDOWN_BUFPUTSL(ob, "</thead>\n");
+}
+
+static void
+rndr_table_body(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
+{
+    if (ob->size) hoedown_buffer_putc(ob, '\n');
+    HOEDOWN_BUFPUTSL(ob, "<tbody>\n");
+    hoedown_buffer_put(ob, content->data, content->size);
+    HOEDOWN_BUFPUTSL(ob, "</tbody>\n");
+}
+
+static void
+rndr_tablerow(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
 	HOEDOWN_BUFPUTSL(ob, "<tr>\n");
-	if (text)
-		hoedown_buffer_put(ob, text->data, text->size);
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</tr>\n");
 }
 
 static void
-rndr_tablecell(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *opaque)
+rndr_tablecell(hoedown_buffer *ob, const hoedown_buffer *content, hoedown_table_flags flags, const hoedown_renderer_data *data)
 {
 	if (flags & HOEDOWN_TABLE_HEADER) {
 		HOEDOWN_BUFPUTSL(ob, "<th");
@@ -498,11 +456,11 @@ rndr_tablecell(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *
 		HOEDOWN_BUFPUTSL(ob, " style=\"text-align: center\">");
 		break;
 
-	case HOEDOWN_TABLE_ALIGN_L:
+	case HOEDOWN_TABLE_ALIGN_LEFT:
 		HOEDOWN_BUFPUTSL(ob, " style=\"text-align: left\">");
 		break;
 
-	case HOEDOWN_TABLE_ALIGN_R:
+	case HOEDOWN_TABLE_ALIGN_RIGHT:
 		HOEDOWN_BUFPUTSL(ob, " style=\"text-align: right\">");
 		break;
 
@@ -510,8 +468,8 @@ rndr_tablecell(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *
 		HOEDOWN_BUFPUTSL(ob, ">");
 	}
 
-	if (text)
-		hoedown_buffer_put(ob, text->data, text->size);
+	if (content)
+		hoedown_buffer_put(ob, content->data, content->size);
 
 	if (flags & HOEDOWN_TABLE_HEADER) {
 		HOEDOWN_BUFPUTSL(ob, "</th>\n");
@@ -521,79 +479,87 @@ rndr_tablecell(hoedown_buffer *ob, const hoedown_buffer *text, int flags, void *
 }
 
 static int
-rndr_superscript(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_superscript(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (!text || !text->size) return 0;
+	if (!content || !content->size) return 0;
 	HOEDOWN_BUFPUTSL(ob, "<sup>");
-	hoedown_buffer_put(ob, text->data, text->size);
+	hoedown_buffer_put(ob, content->data, content->size);
 	HOEDOWN_BUFPUTSL(ob, "</sup>");
 	return 1;
 }
 
 static void
-rndr_normal_text(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_normal_text(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	if (text)
-		escape_html(ob, text->data, text->size);
+	if (content)
+		escape_html(ob, content->data, content->size);
 }
 
 static void
-rndr_footnotes(hoedown_buffer *ob, const hoedown_buffer *text, void *opaque)
+rndr_footnotes(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state = data->opaque;
 
 	if (ob->size) hoedown_buffer_putc(ob, '\n');
 	HOEDOWN_BUFPUTSL(ob, "<div class=\"footnotes\">\n");
 	hoedown_buffer_puts(ob, USE_XHTML(state) ? "<hr/>\n" : "<hr>\n");
 	HOEDOWN_BUFPUTSL(ob, "<ol>\n");
-	
-	if (text)
-		hoedown_buffer_put(ob, text->data, text->size);
-	
+
+	if (content) hoedown_buffer_put(ob, content->data, content->size);
+
 	HOEDOWN_BUFPUTSL(ob, "\n</ol>\n</div>\n");
 }
 
 static void
-rndr_footnote_def(hoedown_buffer *ob, const hoedown_buffer *text, unsigned int num, void *opaque)
+rndr_footnote_def(hoedown_buffer *ob, const hoedown_buffer *content, unsigned int num, const hoedown_renderer_data *data)
 {
 	size_t i = 0;
 	int pfound = 0;
-	
+
 	/* insert anchor at the end of first paragraph block */
-	if (text) {
-		while ((i+3) < text->size) {
-			if (text->data[i++] != '<') continue;
-			if (text->data[i++] != '/') continue;
-			if (text->data[i++] != 'p' && text->data[i] != 'P') continue;
-			if (text->data[i] != '>') continue;
+	if (content) {
+		while ((i+3) < content->size) {
+			if (content->data[i++] != '<') continue;
+			if (content->data[i++] != '/') continue;
+			if (content->data[i++] != 'p' && content->data[i] != 'P') continue;
+			if (content->data[i] != '>') continue;
 			i -= 3;
 			pfound = 1;
 			break;
 		}
 	}
-	
+
 	hoedown_buffer_printf(ob, "\n<li id=\"fn%d\">\n", num);
 	if (pfound) {
-		hoedown_buffer_put(ob, text->data, i);
+		hoedown_buffer_put(ob, content->data, i);
 		hoedown_buffer_printf(ob, "&nbsp;<a href=\"#fnref%d\" rev=\"footnote\">&#8617;</a>", num);
-		hoedown_buffer_put(ob, text->data + i, text->size - i);
-	} else if (text) {
-		hoedown_buffer_put(ob, text->data, text->size);
+		hoedown_buffer_put(ob, content->data + i, content->size - i);
+	} else if (content) {
+		hoedown_buffer_put(ob, content->data, content->size);
 	}
 	HOEDOWN_BUFPUTSL(ob, "</li>\n");
 }
 
 static int
-rndr_footnote_ref(hoedown_buffer *ob, unsigned int num, void *opaque)
+rndr_footnote_ref(hoedown_buffer *ob, unsigned int num, const hoedown_renderer_data *data)
 {
 	hoedown_buffer_printf(ob, "<sup id=\"fnref%d\"><a href=\"#fn%d\" rel=\"footnote\">%d</a></sup>", num, num, num);
 	return 1;
 }
 
-static void
-toc_header(hoedown_buffer *ob, const hoedown_buffer *text, int level, void *opaque)
+static int
+rndr_math(hoedown_buffer *ob, const hoedown_buffer *text, int displaymode, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_buffer_put(ob, (const uint8_t *)(displaymode ? "\\[" : "\\("), 2);
+	escape_html(ob, text->data, text->size);
+	hoedown_buffer_put(ob, (const uint8_t *)(displaymode ? "\\]" : "\\)"), 2);
+	return 1;
+}
+
+static void
+toc_header(hoedown_buffer *ob, const hoedown_buffer *content, int level, const hoedown_renderer_data *data)
+{
+	hoedown_html_renderer_state *state = data->opaque;
 
 	if (level <= state->toc_data.nesting_level) {
 		/* set the level offset if this is the first header
@@ -620,28 +586,34 @@ toc_header(hoedown_buffer *ob, const hoedown_buffer *text, int level, void *opaq
 		}
 
 		hoedown_buffer_printf(ob, "<a href=\"#toc_%d\">", state->toc_data.header_count++);
-		if (text) escape_html(ob, text->data, text->size);
+		if (content) hoedown_buffer_put(ob, content->data, content->size);
 		HOEDOWN_BUFPUTSL(ob, "</a>\n");
 	}
 }
 
 static int
-toc_link(hoedown_buffer *ob, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_buffer *content, void *opaque)
+toc_link(hoedown_buffer *ob, const hoedown_buffer *content, const hoedown_buffer *link, const hoedown_buffer *title, const hoedown_renderer_data *data)
 {
-	if (content && content->size)
-		hoedown_buffer_put(ob, content->data, content->size);
+	if (content && content->size) hoedown_buffer_put(ob, content->data, content->size);
 	return 1;
 }
 
 static void
-toc_finalize(hoedown_buffer *ob, void *opaque)
+toc_finalize(hoedown_buffer *ob, int inline_render, const hoedown_renderer_data *data)
 {
-	rndr_state *state = opaque;
+	hoedown_html_renderer_state *state;
+
+	if (inline_render)
+		return;
+
+	state = data->opaque;
 
 	while (state->toc_data.current_level > 0) {
 		HOEDOWN_BUFPUTSL(ob, "</li>\n</ul>\n");
 		state->toc_data.current_level--;
 	}
+
+	state->toc_data.header_count = 0;
 }
 
 hoedown_renderer *
@@ -649,6 +621,7 @@ hoedown_html_toc_renderer_new(int nesting_level)
 {
 	static const hoedown_renderer cb_default = {
 		NULL,
+
 		NULL,
 		NULL,
 		toc_header,
@@ -661,6 +634,9 @@ hoedown_html_toc_renderer_new(int nesting_level)
 		NULL,
 		NULL,
 		NULL,
+		NULL,
+		NULL,
+		NULL,
 
 		NULL,
 		rndr_codespan,
@@ -672,66 +648,58 @@ hoedown_html_toc_renderer_new(int nesting_level)
 		NULL,
 		NULL,
 		toc_link,
-		NULL,
 		rndr_triple_emphasis,
 		rndr_strikethrough,
 		rndr_superscript,
 		NULL,
-
 		NULL,
 		NULL,
 
 		NULL,
-		toc_finalize,
-		
-		NULL
+		rndr_normal_text,
+
+		NULL,
+		toc_finalize
 	};
 
-	rndr_state       *state;
+	hoedown_html_renderer_state *state;
 	hoedown_renderer *renderer;
 
 	/* Prepare the state pointer */
-	state = malloc(sizeof(rndr_state));
-	if (!state)
-		return NULL;
+	state = hoedown_malloc(sizeof(hoedown_html_renderer_state));
+	memset(state, 0x0, sizeof(hoedown_html_renderer_state));
 
-	memset(state, 0x0, sizeof(rndr_state));
-
-	if (nesting_level > 0) {
-		state->flags |= HOEDOWN_HTML_TOC;
-		state->toc_data.nesting_level = nesting_level;
-	}
+	state->toc_data.nesting_level = nesting_level;
 
 	/* Prepare the renderer */
-	renderer = malloc(sizeof(hoedown_renderer));
-	if (!renderer) {
-		free(state);
-		return NULL;
-	}
-
+	renderer = hoedown_malloc(sizeof(hoedown_renderer));
 	memcpy(renderer, &cb_default, sizeof(hoedown_renderer));
-	
+
 	renderer->opaque = state;
 	return renderer;
 }
 
 hoedown_renderer *
-hoedown_html_renderer_new(unsigned int render_flags, int nesting_level)
+hoedown_html_renderer_new(hoedown_html_flags render_flags, int nesting_level)
 {
 	static const hoedown_renderer cb_default = {
+		NULL,
+
 		rndr_blockcode,
 		rndr_blockquote,
-		rndr_raw_block,
 		rndr_header,
 		rndr_hrule,
 		rndr_list,
 		rndr_listitem,
 		rndr_paragraph,
 		rndr_table,
+		rndr_table_header,
+		rndr_table_body,
 		rndr_tablerow,
 		rndr_tablecell,
 		rndr_footnotes,
 		rndr_footnote_def,
+		rndr_raw_block,
 
 		rndr_autolink,
 		rndr_codespan,
@@ -743,58 +711,37 @@ hoedown_html_renderer_new(unsigned int render_flags, int nesting_level)
 		rndr_image,
 		rndr_linebreak,
 		rndr_link,
-		rndr_raw_html,
 		rndr_triple_emphasis,
 		rndr_strikethrough,
 		rndr_superscript,
 		rndr_footnote_ref,
+		rndr_math,
+		rndr_raw_html,
 
 		NULL,
 		rndr_normal_text,
 
 		NULL,
-		NULL,
-		
 		NULL
 	};
 
-	rndr_state       *state;
+	hoedown_html_renderer_state *state;
 	hoedown_renderer *renderer;
 
 	/* Prepare the state pointer */
-	state = malloc(sizeof(rndr_state));
-	if (!state)
-		return NULL;
-
-	memset(state, 0x0, sizeof(rndr_state));
+	state = hoedown_malloc(sizeof(hoedown_html_renderer_state));
+	memset(state, 0x0, sizeof(hoedown_html_renderer_state));
 
 	state->flags = render_flags;
-
-	if (nesting_level > 0) {
-		state->flags |= HOEDOWN_HTML_TOC;
-		state->toc_data.nesting_level = nesting_level;
-	}
+	state->toc_data.nesting_level = nesting_level;
 
 	/* Prepare the renderer */
-	renderer = malloc(sizeof(hoedown_renderer));
-	if (!renderer) {
-		free(state);
-		return NULL;
-	}
-
+	renderer = hoedown_malloc(sizeof(hoedown_renderer));
 	memcpy(renderer, &cb_default, sizeof(hoedown_renderer));
 
-	if (render_flags & HOEDOWN_HTML_SKIP_IMAGES)
-		renderer->image = NULL;
-
-	if (render_flags & HOEDOWN_HTML_SKIP_LINKS) {
-		renderer->link = NULL;
-		renderer->autolink = NULL;
-	}
-
 	if (render_flags & HOEDOWN_HTML_SKIP_HTML || render_flags & HOEDOWN_HTML_ESCAPE)
 		renderer->blockhtml = NULL;
-	
+
 	renderer->opaque = state;
 	return renderer;
 }
